@@ -1,14 +1,15 @@
 import csv
 import logging
+import warnings
 import os
-import re
+import sys
 from collections import defaultdict
 from itertools import groupby
 
 import numpy as np
 import pandas as pd
 
-from .utils import COLORS, gzopen, merge_intervals
+from .utils import gzopen
 
 try:
     from itertools import ifilterfalse as filterfalse
@@ -96,8 +97,10 @@ def get_traces(data, samples, outliers, distance_threshold, slop):
                         traces[sample]["y"].insert(
                             0, data[sample][index_values[0] - distance_idx]
                         )
-                    except IndexError:
+                    except IndexError as e:
                         # x_values[0] is the first data point
+                        print("index error: %s\nindex_values[0]: %d, distance_idx: %d" % (
+                            e, index_values[0], distance_idx))
                         break
                     extension_length -= (
                         data["x"][index_values[0] - distance_idx + 1]
@@ -164,8 +167,7 @@ def parse_sex_groups(filename, sample_col, sex_col):
             groups[row[sex_col]].append(row[sample_col])
     return groups
 
-
-def normalize_depths(path):
+def normalize_depths(path, sex_chroms, median_window=7):
     filename, ext = os.path.splitext(path)
     if ext == ".gz":
         filename, ext = os.path.splitext(filename)
@@ -176,24 +178,50 @@ def normalize_depths(path):
     df[df.iloc[:, 3:] == 0] = np.nan
     # median values per sample
     global_sample_median = np.nanmedian(df.iloc[:, 3:], axis=0)
-    # normalize each sample
+    # normalize within each sample
     df.iloc[:, 3:] = df.iloc[:, 3:] / global_sample_median
-    # generate output file
-    df.to_csv(path_or_buf=output_bed, sep="\t", na_rep=0.0, index=False)
-    return output_bed
 
+    extras = []
+    for s in sex_chroms:
+        if s.startswith("chr"): extras.append(s[3:])
+        else: extras.append("chr" + s)
+    sex_chroms = sex_chroms + extras
+
+    # median per site
+    autosome = ~np.asarray(df.iloc[:, 0].isin(sex_chroms))
+
+    with warnings.catch_warnings():
+        warnings.filterwarnings('ignore', r'All-NaN (slice|axis) encountered')
+        site_median = np.nanmedian(df.iloc[:, 3:], axis=1)
+
+    with np.errstate(invalid='ignore'):
+        site_median[np.isnan(site_median) | (site_median < 0.03)] = 1
+    # divide autosomes by median at each site so a given block is centered at
+    # middle sample.
+    for i in range(3, df.shape[1]):
+        df.iloc[:, i] = np.where(autosome, df.iloc[:, i] / site_median, df.iloc[:, i])
+        if median_window > 1:
+            df.iloc[:, i] = df.iloc[:, i].rolling(median_window).median() #pd.rolling_median(df.iloc[:, i], median_window)
+        inan = np.asarray(np.isnan(df.iloc[:, i]))
+        df.iloc[inan, i] = 0.0
+    #df.to_csv(path_or_buf=output_bed, sep="\t", na_rep=0.0, index=False,
+    #        compression='gzip',
+    #        float_format="%.2f")
+    return df
+    #return output_bed
 
 def identify_outliers(a, threshold=3.5):
+    a = np.asarray(a, dtype=np.float32)
     med = np.median(a)
-    mad = np.median([np.abs(i - med) for i in a])
+    mad = np.median(np.abs(a - med))
     # https://www.ibm.com/support/knowledgecenter/en/SSEP7J_11.1.0/com.ibm.swg.ba.cognos.ug_ca_dshb.doc/modified_z.html
     if mad == 0:
         meanAD = np.mean(np.abs(a - np.mean(a)))
         divisor = 1.253314 * meanAD
-        modified_z_scores = [(i - med) / divisor for i in a]
+        modified_z_scores = (a - med) / divisor
     else:
         divisor = 1.4826 * mad
-        modified_z_scores = [(i - med) / divisor for i in a]
+        modified_z_scores = (a - med) / divisor
     return np.where(np.abs(modified_z_scores) > threshold)
 
 
@@ -202,7 +230,9 @@ def add_roc_traces(path, traces, exclude):
     df = pd.read_csv(path, sep="\t", low_memory=False)
     n_bins = 150
     x_max = 2.5
-    x = list(np.linspace(0, x_max, n_bins))
+    traces["roc"]["x_coords"] = [
+        round(i, 2) for i in list(np.linspace(0, x_max, n_bins))
+    ]
 
     for chrom, data in df.groupby(df.columns[0]):
         chrom = str(chrom)
@@ -210,29 +240,20 @@ def add_roc_traces(path, traces, exclude):
         if exclude.findall(chrom):
             continue
 
-        chrom = chrom.lstrip("chr")
+        chrom = chrom[3:] if chrom.startswith("chr") else chrom
 
         # pre-normalized data
         arr = np.asarray(data.iloc[:, 3:])
-        traces["roc"][chrom] = list()
+        traces["roc"][chrom] = dict()
 
         for i in range(0, arr.shape[1]):
             # get counts across our x-range of bins
             counts, _ = np.histogram(arr[:, i], bins=n_bins, range=(0, x_max))
             # decreasing order of the cumulative sum across the bins
             sums = counts[::-1].cumsum()[::-1]
-            # normalize to y_max to 1
+            # normalize to y_max of 1
             sums = list(sums.astype(float) / max(1, sums[0]))
-
-            trace = dict(
-                x=x,
-                y=sums,
-                hoverinfo="text",
-                mode="lines",
-                text=df.columns[i + 3],
-                marker={"color": "rgba(108,117,125,0.2)"},
-            )
-            traces["roc"][chrom].append(trace)
+            traces["roc"][chrom][df.columns[i + 3]] = [round(i, 2) for i in sums]
     return traces
 
 
@@ -248,13 +269,11 @@ def parse_bed(
     slop=500000,
     min_samples=8,
     skip_norm=False,
+    window=1
 ):
     bed_traces = dict()
-    exclusions = re.compile(exclude)
     # chromosomes, in order of appearance
     chroms = list()
-    samples = list()
-    sample_trace_color = dict()
 
     sex_chroms = [i.strip("chr") for i in sex_chroms.split(",")]
 
@@ -262,22 +281,32 @@ def parse_bed(
     if ped:
         groups = parse_sex_groups(ped, sample_col, sex_col)
 
-    if not skip_norm:
-        path = normalize_depths(path)
+    if skip_norm:
+        df = pd.read_csv(path, sep="\t", low_memory=False)
+    else:
+        df = normalize_depths(path, sex_chroms, median_window=window)
+
+    if True: # temporary to get sane diff.
+        header = list(df.columns)
+        samples = header[3:]
+        if groups:
+            valid = validate_samples(samples, groups)
+            if not valid:
+                logger.critical(
+                    "sample ID mismatches exist between ped and bed"
+                )
+                sys.exit(1)
+        for chr, entries in df.groupby(header[0], as_index=False, sort=False):
 
-    with gzopen(path) as fh:
-        header = fh.readline().strip().split("\t")
-        fh.seek(0)
-        reader = csv.DictReader(fh, delimiter="\t")
-        for chr, entries in groupby(reader, key=lambda i: i[header[0]]):
             # apply exclusions
             if exclude.findall(chr):
+                logger.debug("excluding chromosome: %s" % chr)
                 continue
 
             data = defaultdict(list)
             bounds = dict(upper=[], lower=[])
             outliers = defaultdict(list)
-            chrom = chr.lstrip("chr")
+            chrom = chr[3:] if chr.startswith("chr") else chr
             chroms.append(chrom)
 
             # capture plot area and outlier traces
@@ -285,21 +314,14 @@ def parse_bed(
             if chrom in sex_chroms and groups:
                 sample_groups = groups
 
-            for x_index, row in enumerate(entries):
-                if not samples:
-                    samples = [i for i in sorted(row.keys()) if i not in header[:3]]
-                    # plot outlier traces across chromosomes with the same color for a given sample
-                    for sample_index, sample in enumerate(samples):
-                        sample_trace_color[sample] = COLORS[sample_index % len(COLORS)]
-                    if groups:
-                        valid = validate_samples(samples, groups)
-                        if not valid:
-                            logger.critical(
-                                "sample ID mismatches exist between ped and bed"
-                            )
-                            sys.exit(1)
-                if sample_groups is None:
-                    sample_groups = {"gid": samples}
+            all_points = sample_groups is None
+
+            if sample_groups is None:
+                sample_groups = {"gid": samples}
+
+            x_index = -1
+            for not_x_index, row in entries.iterrows():
+                x_index += 1
 
                 x_value = int(row[header[1]])
                 data["x"].append(x_value)
@@ -307,16 +329,21 @@ def parse_bed(
                 for group_index, (gid, samples_of_group) in enumerate(
                     sample_groups.items()
                 ):
+                    # adds area traces where groups are present (chrs X and Y)
                     if len(bounds["upper"]) == group_index:
                         bounds["upper"].append([])
                         bounds["lower"].append([])
-                    sample_values = []
-                    for sample in samples_of_group:
-                        v = float(row[sample])
-                        if v > 3:
-                            v = 3
-                        data[sample].append(v)
-                        sample_values.append(v)
+                    if all_points:
+                        sample_values = np.minimum(3, np.asarray(row[3:],
+                            dtype=np.float32))
+                        for i, s in enumerate(samples):
+                            data[s].append(float(sample_values[i]))
+                    else:
+                        sample_values = []
+                        for sample in samples_of_group:
+                            v = min(3.0, float(row[sample]))
+                            data[sample].append(v)
+                            sample_values.append(v)
 
                     # skip finding outliers for few samples
                     if len(samples) <= min_samples:
@@ -329,17 +356,21 @@ def parse_bed(
 
                     # skip running test if everything is the same
                     if len(set(sample_values)) == 1:
-                        bounds["upper"][group_index].append(sample_values[0])
-                        bounds["lower"][group_index].append(sample_values[0])
+                        bounds["upper"][group_index].append(float(sample_values[0]))
+                        bounds["lower"][group_index].append(float(sample_values[0]))
                     else:
                         # indexes of passing values
                         passing = identify_outliers(sample_values, z_threshold)[0]
                         # remove those indexes from the list
+                        # sometimes we get a list (x, y) for others we get a
+                        # numpy array
+                        if len(passing) > 0 and not isinstance(sample_values, list):
+                            sample_values = sample_values.tolist()
                         for j in sorted(passing, reverse=True):
                             sample_values.pop(j)
                         # from remaining, grab upper and lower bounds
-                        upper = max(sample_values)
-                        lower = min(sample_values)
+                        upper = float(max(sample_values))
+                        lower = float(min(sample_values))
                         bounds["upper"][group_index].append(upper)
                         bounds["lower"][group_index].append(lower)
                         required_deviation_from_bounds = 0.3
@@ -355,62 +386,49 @@ def parse_bed(
 
             # update the outlier traces
             traces = get_traces(data, samples, outliers, distance_threshold, slop)
+            json_output = dict(upper=[], lower=[], coords=data["x"], samples=[])
 
-            json_output = []
-
-            marker_color = "rgba(108,117,125,0.1)"
-            fill_color = "rgba(108,117,125,0.3)"
             # add the area traces
             for trace_index in range(len(bounds["upper"])):
                 for bound in ["lower", "upper"]:
-                    trace = dict(
-                        x=data["x"],
-                        y=bounds[bound][trace_index],
-                        fill="none",
-                        type="scatter",
-                        mode="lines",
-                        hoverinfo="none",
-                        marker={"color": marker_color},
+                    json_output[bound].append(
+                        [round(i, 2) for i in bounds[bound][trace_index]]
                     )
-                    if bound == "upper":
-                        trace["fill"] = "tonexty"
-                        trace["fillcolor"] = fill_color
-                    json_output.append(trace)
             # add the sample traces for the outlier plots atop area traces
             len_traces = 0
             for sample, trace_data in traces.items():
-                trace = dict(
-                    x=trace_data["x"],
-                    y=trace_data["y"],
-                    text=sample,
-                    connectgaps=False,
-                    hoverinfo="text",
-                    mode="lines",
-                    name="significant",
-                    # include color as primary colors occupied by area traces
-                    marker={"width": 1, "color": sample_trace_color[sample]},
+                if not trace_data["x"]:
+                    continue
+                # y data may be gapped (string separated floats)
+                y_data = list()
+                for v in trace_data["y"]:
+                    try:
+                        y_data.append(round(v, 2))
+                    except TypeError:
+                        y_data.append(v)
+                json_output["samples"].append(
+                    {"name": sample, "x": trace_data["x"], "y": y_data}
                 )
-                if trace_data["x"]:
-                    len_traces += 1
-                json_output.append(trace)
+                len_traces += 1
 
             bed_traces[chrom] = json_output
             logger.info("plotting %d traces on chrom %s" % (len_traces, chr))
 
     bed_traces["chromosomes"] = chroms
-    bed_traces["sex_chroms"] = sex_chroms
+    bed_traces["sample_list"] = samples
+    # bed_traces["sex_chroms"] = sex_chroms
 
     # pass the bed or normed bed
     add_roc_traces(path, bed_traces, exclude)
 
-    return bed_traces, samples
+    return bed_traces
 
 
-def parse_bed_track(path, traces, exclude, y_offset=-0.15, track_color="#444"):
+def parse_bed_track(path, traces, exclude):
     """
     parse a bed file, placing lines per region
     """
-    trace_name = os.path.basename(path)
+    trace_name = os.path.basename(path).partition(".bed")[0]
 
     with gzopen(path) as fh:
         cleaned = filterfalse(lambda i: i[0] == "#", fh)
@@ -423,6 +441,12 @@ def parse_bed_track(path, traces, exclude, y_offset=-0.15, track_color="#444"):
                 continue
             if chrom not in traces:
                 continue
+
+            if not "annotations" in traces[chrom]:
+                traces[chrom]["annotations"] = {"bed": []}
+            if not "bed" in traces[chrom]["annotations"]:
+                traces[chrom]["annotations"]["bed"] = list()
+
             regions = list()
             for line in entries:
                 if line.startswith("#"):
@@ -435,42 +459,8 @@ def parse_bed_track(path, traces, exclude, y_offset=-0.15, track_color="#444"):
                     name = toks[3]
                 except IndexError:
                     name = ""
-                regions.append([start, end, [name]])
-            if regions:
-                merged_regions = merge_intervals(regions)
-                # update list to semi-colon delimited string
-                for interval in merged_regions:
-                    interval[2] = ";".join(set(interval[2]))
-                    if len(interval[2]) > 55:
-                        interval[2] = interval[2][0:55] + "..."
+                regions.append([start, end, name])
+
+            traces[chrom]["annotations"]["bed"].append([trace_name, regions])
 
-                x_values = list()
-                y_values = list()
-                text_values = list()
-                for interval in merged_regions:
-                    # start
-                    x_values.append(interval[0])
-                    # end
-                    x_values.append(interval[1])
-                    # gap
-                    x_values.append("")
-                    y_values.append(y_offset)
-                    y_values.append(y_offset)
-                    y_values.append("")
-                    text_values.append(interval[2])
-                    text_values.append(interval[2])
-                    text_values.append("")
-                trace = dict(
-                    x=x_values,
-                    y=y_values,
-                    text=text_values,
-                    type="scattergl",
-                    name=trace_name,
-                    tracktype="bed",
-                    connectgaps=False,
-                    # hoverinfo="text",
-                    showlegend=False,
-                    line={"width": 10, "color": track_color},
-                )
-                traces[chrom].append(trace)
     return traces
